/*
 * Copyright (c) 1995, the EUROPAGATE consortium (see below).
 *
 * The EUROPAGATE consortium members are:
 *
 *    University College Dublin
 *    Danmarks Teknologiske Videnscenter
 *    An Chomhairle Leabharlanna
 *    Consejo Superior de Investigaciones Cientificas
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation, in whole or in part, for any purpose, is hereby granted,
 * provided that:
 *
 * 1. This copyright and permission notice appear in all copies of the
 * software and its documentation. Notices of copyright or attribution
 * which appear at the beginning of any file must remain unchanged.
 *
 * 2. The names of EUROPAGATE or the project partners may not be used to
 * endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * 3. Users of this software (implementors and gateway operators) agree to
 * inform the EUROPAGATE consortium of their use of the software. This
 * information will be used to evaluate the EUROPAGATE project and the
 * software, and to plan further developments. The consortium may use
 * the information in later publications.
 * 
 * 4. Users of this software agree to make their best efforts, when
 * documenting their use of the software, to acknowledge the EUROPAGATE
 * consortium, and the role played by the software in their work.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED, OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL THE EUROPAGATE CONSORTIUM OR ITS MEMBERS BE LIABLE
 * FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF
 * ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND
 * ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE
 * USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */
/*
 * FML interpreter. Europagate, 1995
 *
 * $Log: fmlarit.c,v $
 * Revision 1.4  1995/05/16 09:39:32  adam
 * LICENSE.
 *
 * Revision 1.3  1995/02/23  08:32:04  adam
 * Changed header.
 *
 * Revision 1.1  1995/02/09  14:33:36  adam
 * Split source fml.c and define relevant build-in functions in separate
 * files. New operators mult, div, not, llen implemented.
 *
 */

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

#include "fmlp.h"

static struct fml_node *fml_exec_plus (Fml fml, struct fml_node *l,
                                       struct fml_node *r)
{
    int left_val, right_val;

    fml_lr_values (fml, l, &left_val, r, &right_val);
    return fml_mk_node_val (fml, left_val + right_val);
}

static struct fml_node *fml_exec_minus (Fml fml, struct fml_node *l,
                                       struct fml_node *r)
{
    int left_val, right_val;

    fml_lr_values (fml, l, &left_val, r, &right_val);
    return fml_mk_node_val (fml, left_val - right_val);
}

static struct fml_node *fml_exec_mult (Fml fml, struct fml_node *l,
                                       struct fml_node *r)
{
    int left_val, right_val;

    fml_lr_values (fml, l, &left_val, r, &right_val);
    return fml_mk_node_val (fml, left_val * right_val);
}

static struct fml_node *fml_exec_div (Fml fml, struct fml_node *l,
                                       struct fml_node *r)
{
    int left_val, right_val;

    fml_lr_values (fml, l, &left_val, r, &right_val);
    return fml_mk_node_val (fml, left_val / right_val);
}

static struct fml_node *fml_exec_incr (Fml fml, struct fml_node **lp, 
                                       struct token *tp)
{
    struct fml_node *fn = NULL;
    struct fml_sym_info *info;
    fml_cmd_lex (lp, tp);
    if (tp->kind == 'e')
    {
        info = fml_sym_lookup (fml->sym_tab, tp->tokenbuf);
        assert (info);
        if (info->kind == FML_VAR && info->body && info->body->is_atom)
        {
            char arg[128];
            int val;
            
            val = fml_atom_val (info->body->p[0]);
            fml_node_delete (fml, info->body);
            sprintf (arg, "%d", val+1);
            info->body = fn = fml_node_alloc (fml);
            fn->is_atom = 1;
            fn->p[0] = fml_atom_alloc (fml, arg);
        }
    }
    fml_cmd_lex (lp, tp);
    return NULL;
}

static struct fml_node *fml_exec_decr (Fml fml, struct fml_node **lp, 
                                       struct token *tp)
{
    struct fml_node *fn = NULL;
    struct fml_sym_info *info;
    fml_cmd_lex (lp, tp);
    if (tp->kind == 'e')
    {
        info = fml_sym_lookup (fml->sym_tab, tp->tokenbuf);
        assert (info);
        if (info->kind == FML_VAR && info->body && info->body->is_atom)
        {
            char arg[128];
            int val;
            
            val = fml_atom_val (info->body->p[0]);
            sprintf (arg, "%d", val-1);
            info->body = fn = fml_node_alloc (fml);
            fn->is_atom = 1;
            fn->p[0] = fml_atom_alloc (fml, arg);
        }
    }
    fml_cmd_lex (lp, tp);
    return NULL;
}

void fml_arit_init (Fml fml)
{
    struct fml_sym_info *sym_info;

    sym_info = fml_sym_add (fml->sym_tab, "plus");
    sym_info->kind = FML_CBINARY;
    sym_info->binary = fml_exec_plus;
    sym_info = fml_sym_add (fml->sym_tab, "minus");
    sym_info->kind = FML_CBINARY;
    sym_info->binary = fml_exec_minus;

    sym_info = fml_sym_add (fml->sym_tab, "mult");
    sym_info->kind = FML_CBINARY;
    sym_info->binary = fml_exec_mult;
    sym_info = fml_sym_add (fml->sym_tab, "div");
    sym_info->kind = FML_CBINARY;
    sym_info->binary = fml_exec_div;

    sym_info = fml_sym_add (fml->sym_tab, "incr");
    sym_info->kind = FML_CPREFIX;
    sym_info->prefix = fml_exec_incr;
    sym_info = fml_sym_add (fml->sym_tab, "decr");
    sym_info->kind = FML_CPREFIX;
    sym_info->prefix = fml_exec_decr;
}
