/*
 * Copyright (c) 1995, the EUROPAGATE consortium (see below).
 *
 * The EUROPAGATE consortium members are:
 *
 *    University College Dublin
 *    Danmarks Teknologiske Videnscenter
 *    An Chomhairle Leabharlanna
 *    Consejo Superior de Investigaciones Cientificas
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation, in whole or in part, for any purpose, is hereby granted,
 * provided that:
 *
 * 1. This copyright and permission notice appear in all copies of the
 * software and its documentation. Notices of copyright or attribution
 * which appear at the beginning of any file must remain unchanged.
 *
 * 2. The names of EUROPAGATE or the project partners may not be used to
 * endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * 3. Users of this software (implementors and gateway operators) agree to
 * inform the EUROPAGATE consortium of their use of the software. This
 * information will be used to evaluate the EUROPAGATE project and the
 * software, and to plan further developments. The consortium may use
 * the information in later publications.
 * 
 * 4. Users of this software agree to make their best efforts, when
 * documenting their use of the software, to acknowledge the EUROPAGATE
 * consortium, and the role played by the software in their work.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED, OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL THE EUROPAGATE CONSORTIUM OR ITS MEMBERS BE LIABLE
 * FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF
 * ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND
 * ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE
 * USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */
/*
 * Iso2709 record management
 *
 * Europagate, 1994-1995.
 *
 * $Log: iso27dis.c,v $
 * Revision 1.5  1995/05/16 09:40:54  adam
 * LICENSE.
 *
 * Revision 1.4  1995/03/29  11:44:29  adam
 * New functions: iso2709_a_.. for record manipulation.
 *
 * Revision 1.3  1995/02/22  21:32:36  adam
 * Changed header.
 *
 * Revision 1.1  1995/02/10  17:05:18  adam
 * New function iso2709_display to display MARC records in a
 * line-by-line format. The iso2709_cvt function no longer
 * prints the record to stderr.
 *
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>
#include <ctype.h>

#include <iso2709.h>

void iso2709_display (Iso2709Rec rec, FILE *out)
{
    Iso2709Anchor a;
    char *tag;
    char *indicator;
    char *identifier;
    char *data;
    
    a = iso2709_a_mk (rec);
    do
    {
        if (!iso2709_a_info_line (a, &tag, &indicator))
            break;
        fprintf (out, "%s", tag);
        if (indicator)
            fprintf (out, " %s", indicator);
        do
        {
            iso2709_a_info_field (a, NULL, NULL, &identifier, &data);
            if (identifier)
                fprintf (out, " $%s", identifier);
            fprintf (out, " %s", data);
        } while (iso2709_a_next_field (a));
        fprintf (out, "\n");
    } while (iso2709_a_next_line(a));
}

